"""Plotting methods for haplotype probabilities

"""

from collections import Counter
from typing import Dict, List, Tuple

import numpy as np
import matplotlib.pyplot as plt


def plot_haplotypes(
        proba_haplotypes: Dict[Tuple[str], float],
        thres=1e-3,
        title=r"Haplotype relative frequencies (area $\sim$ p)",
        **kwargs
):
    """Visualize non-zero probabilities as a Hinton diagram

    """

    (haplotypes, ps) = zip(
        *sorted(filter(lambda xs: xs[1] > thres, proba_haplotypes.items()))
    )

    (fig, ax) = plt.subplots(**kwargs)
    ax = hinton(ps, ax=ax)
    ax.set_xticks(np.arange(len(ps)))
    ax.set_xticklabels(
        list(map(lambda x: "".join(x), haplotypes)),
        rotation=30,
        ha="right"
    )
    ax.set_title(title)

    return fig


def plot_phenotypes(phenotypes, title="Phenotype counts", **kwargs):
    """Visualize phenotype data as a historgram

    """

    (labels, value_counts) = zip(*Counter(phenotypes).items())

    (fig, ax) = plt.subplots(**kwargs)

    ax.bar(np.arange(len(value_counts)), value_counts, align="center")
    ax.set_xticks(np.arange(len(value_counts)))
    ax.set_xticklabels(
        list(map(lambda xs: "(" + ", ".join(xs) + ")", labels)),
        rotation=30,
        ha="right"
    )
    ax.set_title("Phenotype counts")
    ax.set_ylabel("Value counts")

    fig.tight_layout()

    return fig


def hinton(ps: List[float], ax):
    """Visualize an array as a Hinton diagram

    """

    ax.patch.set_facecolor('gray')
    ax.set_aspect('equal', 'box')
    ax.xaxis.set_major_locator(plt.NullLocator())
    ax.yaxis.set_major_locator(plt.NullLocator())

    for (i, p) in enumerate(ps):
        size = np.sqrt(abs(p))
        ax.add_patch(
            plt.Rectangle(
                [i - size / 2, -size / 2],
                size,
                size,
                facecolor="white",
                edgecolor="white"
            )
        )

    ax.autoscale_view()

    return ax



"""HaploPy package

"""

import importlib.metadata

from .datautils import *
from .multinomial import *

try:
    from .plot import *
except ImportError:
    # Allow import plot fail in environments where matplotlib is not installed.
    pass


__version__ = importlib.metadata.version("haplopy")



"""Data utils for HaploPy

Terminology
-----------

haplotype : A sequence of nucleotides.
            Example: ("a", "b")

diplotype : A pair of haplotypes.
            Example: (("a", "b"), ("A", "B"))

phenotype : A sequence of nucleotide pairs with unspecified diplotype.
            Example: ("Aa", "Bb")

genotype  : An item in the phenotype.
            Example: "Aa"

"""

from collections import Counter
from functools import reduce
import itertools
from operator import itemgetter
import re
from typing import Dict, List, Set, Tuple

import numpy as np
from scipy.sparse import dok_matrix


#
# Generic
#


def compose2(f, g):
    return lambda *args, **kwargs: f(g(*args, **kwargs))


def dmax(x: dict):
    """Dictionary's (argmax, max) key-value pair

    Note that this is a pure function. From Python 3.7 onwards, a dictionary is
    ordered by construction. Thus, in case of multiple keys of same numeric
    value, the result only depends on inputs initial order.

    Examples
    --------
    >>> dmax({"a": 1, "b": np.NaN, "c": 4})
    ("c", 4)

    """
    return max(
        filter(
            compose2(np.isfinite, itemgetter(1)),
            x.items()
        ),
        key=itemgetter(1),
        default=(None, None)
    )


#
# Haplotypes
#


def match(haplotype, haplotypes) -> List[Tuple[str]]:
    """Find haplotypes that match a given haplotype

    Examples
    --------
    >>> match(("a", "."), [("a", "b"), ("a", "B"), ("A", "B")])
    [("a", "b"), ("a", "B")]

    """


    def validate(s: str):
        valid = bool(re.compile(r"[a-zA-Z0-9\.]{2}").match(s))
        if not valid:
            raise ValueError(
                (
                    "Illegal characters in {} for pattern matching."
                    "Only a-z, A-Z and . are allowed."
                ).format(tuple(s))
            )
        return s


    return list(map(
        # Transform joint 2-strings back to tuples
        tuple,
        # Filter out 2-strings that don't match with the pattern
        filter(
            # Transform test haplotype into a regex pattern
            re.compile(validate("".join(haplotype))).match,
            # Join reference haplotypes into a list of 2-strings
            map(lambda h: "".join(h), haplotypes)
        )
    ))


#
# Diplotypes
#


def unphase(diplotype) -> Tuple[str]:
    """The most basic operation diplotype to phenotype mapping

    Examples
    --------
    >>> unphase((("A", "T", "G"), ("A", "A", "C")))
    ("AA", "AT", "GC")

    """
    return tuple("".join(sorted(snp)) for snp in zip(*diplotype))


def fill(diplotype, haplotypes) -> List[Tuple[Tuple[str]]]:
    """Attempt filling in missing values inside a diplotype

    Uses regular expressions to fill in missing SNPs with admissible values
    that are present in `haplotypes`.

    Leaves unmatchable patterns as they are.

    Examples
    --------
    >>> fill(
    ...     (("A", "."), ("a", "b")),
    ...     [("A", "B"), ("A", "b"), ("a", "B"), ("a", "b")]
    ... )
    [(("A", "B"), ("a", "b")), (("A", "b"), ("a", "b"))]

    """

    def expand(h):
        e = match(h, haplotypes)
        return (e if e else [h])

    def add_new(ds, d):
        return ds + [d] if not (d in ds or d[::-1] in ds) else ds

    return reduce(
        # Select unique (up to order switch) pairs
        add_new,
        # Cartesian product of all matching diplotypes
        itertools.product(*map(expand, diplotype)),
        []
    )


#
# Phenotypes
#


def count_distinct(phenotypes):
    """Count distinct phenotypes in a list

    Importantly, standardizes a dataset regarding differently ordered genotype
    strings within a phenotype.

    Examples
    --------
    >>> count_distinct([("aA", "bB"), ("Aa", "Bb")])
    Counter({('Aa', 'Bb'): 2})

    """

    # At the moment we have decided not to implement a tailored data type
    # for phenotypes, we need to resort to hacky sorting and string joins
    # to avoid misclassifying equivalent phenotypes with permuted genotypes.
    def sort_genotypes(phenotype):
        return tuple("".join(sorted(g)) for g in phenotype)

    return Counter(map(sort_genotypes, phenotypes))


def factorize(phenotype) -> List[Tuple[str]]:
    """List admissible diplotypes

    Examples
    --------
    >>> factorize(("Aa", "BB", "Cc"))
    [(('A', 'B', 'C'), ('a', 'B', 'c')), (('A', 'B', 'c'), ('a', 'B', 'C'))]

    """
    factors = list(itertools.product(*[
        # NOTE: Without sorting the function wouldn't be pure: set of string
        # doesn't have fixed order.
        sorted(set(locus)) for locus in phenotype
    ]))
    half = len(factors) // 2
    return (
        [(factors[0], factors[0])] if half == 0 else
        list(zip(factors[:half], factors[half:][::-1]))
    )


def factorize_fill(phenotype, haplotypes) -> List[Tuple[Tuple[str]]]:
    """Factorize to diplotypes with missing value filling

    Examples
    --------
    >>> factorize_fill(
    ...     ("Aa", "B."),
    ...     [("a", "b"), ("A", "b"), ("a", "B"), ("A", "B")]
    ... )
    [
        (('A', 'B'), ('a', 'B')),
        (('A', 'B'), ('a', 'b')),
        (('A', 'b'), ('a', 'B'))
    ]

    """
    # NOTE: Without sorting the function wouldn't be pure: set of string
    # doesn't have fixed order.
    return sorted(reduce(
        lambda ds, d: ds.union(
            map(
                lambda x: tuple(sorted(x)),
                fill(d, haplotypes)
            )
        ),
        factorize(phenotype),
        set()
    ))


def find_admissible_haplotypes(counter) -> List[Tuple[str]]:
    """List parent haplotypes

    Examples
    --------
    >>> from collections import Counter
    >>> find_admissible_diplotypes(Counter({
    ...     ("AA", "BB"): 2,
    ...     ("aa", "bb"): 1
    ... }))
    [("A", "B"), ("a", "b")]

    """
    # Sort the reduced set to fix output ordering.
    # In this way, we get a pure function.
    return sorted(
        reduce(
            lambda acc, x: acc.union(
                # Set of admissible haplotypes for single phenotype
                set(itertools.product(*map(set, x)))
            ),
            counter,
            set()
        )
    )


def build_diplotype_representation(counter, haplotypes) -> List[List[Tuple[int]]]:
    """Phenotype multiplicity and parent diplotype expansion

    Parameters
    ----------
    counter : collections.Counter
        Counts of distinct phenotypes, that is, "intra-genotype" string ordering
        doesn't matter.

    Returns
    -------
    representation : List[List[Tuple[int]]]
        Each item corresponds to the element in `counter` with same index.
        The item is a list of index pairs. Each index points to an element
        in `parent_haplotypes`, and the pair stands for an admissible parent
        haplotype couple.

    Examples
    --------
    >>> from collections import Counter
    >>> counter = Counter({("Aa", "Bb"): 2, ("aa", "bb"): 1})
    >>> haplotypes = [('a', 'b'), ('A', 'B'), ('a', 'B'), ('A', 'b')]
    >>> build_diplotype_representation(counter, haplotypes)
    [[(1, 0), (3, 2)], [(0, 0)]]

    """

    def factorize_to_index(phenotype):
        return [
            (haplotypes.index(x), haplotypes.index(y))
            for (x, y) in factorize(phenotype)
        ]

    return list(map(factorize_to_index, counter))


def build_diplotype_matrix(
        diplotype_representation: List[List[Tuple[int]]],
        haplotypes: List[Tuple[str]]
):
    """Haplotype multiplicity in a 'N haplotypes' * 'M diplotypes' matrix

    Indicates out how many times haplotype n is present in diplotype m

    Examples
    --------
    >>> from collections import Counter
    >>> counter = Counter({("Aa", "Bb"): 2, ("aa", "bb"): 1})
    >>> haplotypes = [('a', 'b'), ('A', 'B'), ('a', 'B'), ('A', 'b')]
    >>> drepr = build_diplotype_representation(counter, haplotypes)
    >>> build_diplotype_matrix(drepr, haplotypes).toarray()
    array([[1, 0, 2],
           [1, 0, 0],
           [0, 1, 0],
           [0, 1, 0]])


    """

    diplotypes = reduce(lambda x, y: x + y, diplotype_representation, [])
    matrix = dok_matrix(
        (len(haplotypes), len(diplotypes)),
        dtype=int
    )

    # Populate matrix
    for (i, diplotype) in enumerate(diplotypes):
        matrix[diplotype[0], i] += 1
        matrix[diplotype[1], i] += 1

    return matrix



"""Multinomial distribution haplotype frequency inference model

TODO: Random generate EM initial values

"""


from __future__ import annotations

from functools import reduce
import json
import logging
from typing import Dict, List, Tuple

import numpy as np

from haplopy import datautils


def log_multinomial(*ns: int):
    """Logarithm of the multinomial coefficient

        (n1 + n2 + ... + nk)!
        ---------------------
        n1!  n2!  ...   nk!

    log(n!) = log(n) + log(n-1) + log(n-2) + ... + log(1)

    """

    def log_factorial(n):
        return np.log(np.arange(1, n + 1)).sum()

    return log_factorial(sum(ns)) - sum(map(log_factorial, ns))


def expectation_maximization(
        phenotypes: List[Tuple[str]],
        proba_haplotypes: Dict[Tuple[str], float]=None,
        max_iter: int=100,
        tol: float=1.0e-12
):
    """Expectation Maximization algorithm for haplotype probabilities

    Tries to iteratively find the Maximum Likelihood estimate.

    Philosophically, builds upon the hypotheses of Hardy-Weinberg equilibrium
    and random mating.

    Parameters
    ----------

    phenotypes : List[Tuple[str]]
        Observation data; list of phenotypes
    proba_haplotypes : Dict[Tuple[str], float]
        Initial haplotype probabilities
    max_iter : int
        Maximum number of iterations
    tol : float
        Stopping criterion with respect to Euclidean norm of probabilities vector

    """

    N = len(phenotypes)

    # If haplotype probabilities are not given, a uniform initial probability
    # is assumed, and only admissible haplotypes are considered
    if proba_haplotypes is None:
        haplotypes = datautils.find_admissible_haplotypes(phenotypes)
        Nh = len(haplotypes)
        probas = np.ones(Nh) / Nh
    else:
        (haplotypes, probas) = zip(*proba_haplotypes.items())

    counter = datautils.count_distinct(phenotypes)
    diplotype_representation = datautils.build_diplotype_representation(
        counter, haplotypes
    )

    # Diplotype count matrix for faster multiplication
    diplotype_matrix = datautils.build_diplotype_matrix(
        diplotype_representation, haplotypes
    )

    #
    # Separation into expectation and maximization
    #
    # This is not a mathematically accurate separation
    # but loosely based on [1]. The update formulae are non-trivial to
    # derive. A clear mathematical proof is found, e.g., in [2].
    #
    # [1] J. Polanska, The EM algorithm and its implementation for the
    #     estimation of frequencies of SNP-haplotypes, Int. J. Appl. Math.
    #     Comput. Sci., 2003, Vol. 13, No. 3, 419-429.
    #
    # [2] Mikko Koivisto, 582673 Computational Genotype Analysis (lecture 2
    #     notes), Uni. Helsinki
    #

    def expectation(probas: np.ndarray):
        """Calculate probability for each pair and evaluate log-likelihood

        """

        def calculate(inds):
            return np.array([
                2 ** (i != j) * probas[i] * probas[j] for (i, j) in inds
            ])

        Ps_raws = [calculate(inds) for inds in diplotype_representation]
        Ps_units = np.hstack([
            Ps * n / Ps.sum() / N for (Ps, n) in zip(Ps_raws, counter.values())
        ])
        log_likelihood = sum([
            n * np.log(Ps.sum()) for (Ps, n) in zip(Ps_raws, counter.values())
        ]) + log_multinomial(*counter.values())

        return (Ps_units, log_likelihood)

    def maximization(Ps):
        """Calculate next estimates of haplotype probabilities

        """
        return 0.5 * diplotype_matrix.dot(Ps)

    n_iter = 0
    step = np.inf
    logging.info("Start EM algorithm")
    while (n_iter <= max_iter) and (step > tol):
        (Ps, log_likelihood) = expectation(probas)
        delta = maximization(Ps) - probas
        step = np.linalg.norm(delta)
        probas = probas + delta
        logging.info(
            "Iteration {0} | log(L) = {1:.4e} | step = {2: .4e}".format(
                n_iter, log_likelihood, step
            )
        )
        n_iter += 1

    proba_haplotypes = dict(zip(haplotypes, probas))

    return (proba_haplotypes, log_likelihood)


class Model():
    """Haplotype multinomial distribution estimator

    Examples
    --------
    >>> model = Model({("a", "b"): 0.5, ("A", "B"): 0.5})
    >>> phenotypes = model.random(10)
    >>> model_fitted = Model.fit(phenotypes)
    >>> model_fitted.proba_haplotypes
    {('A', 'B'): 0.6,
     ('A', 'b'): 0.0,
     ('a', 'B'): 0.0,
     ('a', 'b'): 0.4}


    """

    def __init__(self, proba_haplotypes: Dict[Tuple[str], float]):
        (haplotypes, probas) = zip(*proba_haplotypes.items())
        assert abs(sum(probas) - 1) < 1e-8, "Probabilities must sum to one"
        self.proba_haplotypes = proba_haplotypes
        return

    def random(self, n_obs: int) -> List[Tuple[str]]:
        """Random generate phenotypes

        """
        (haplotypes, probas) = zip(*self.proba_haplotypes.items())
        parent_inds = np.dot(
            np.random.multinomial(1, probas, 2 * n_obs),
            np.arange(len(probas))
        ).reshape(n_obs, 2)
        return [
            tuple(
                "".join(sorted(snp)) for snp in zip(haplotypes[i], haplotypes[j])
            )
            for (i, j) in parent_inds
        ]

    @classmethod
    def fit(cls, phenotypes: Tuple[str], **kwargs) -> Model:
        """Fit maximum likelihood haplotype probabilities using EM algorithm

        Implemented as a classmethod so that we can use given phenotypes
        to initialize the model conveniently.

        """
        (proba_haplotypes, _) = expectation_maximization(phenotypes, **kwargs)
        return cls(proba_haplotypes=proba_haplotypes)

    def calculate_proba_diplotypes(
            self,
            phenotype: Tuple[str],
            fill_proba: float=np.NaN,
    ) -> Dict[Tuple[int], float]:
        """Calculate admissible diplotypes' conditional probabilities

        Examples
        --------
        >>> model = Model({("a", "b"): 0.2, ("A", "B"): 0.5, ("a", "B"): 0.3})
        >>> model.calculate_proba_diplotypes(("Aa", "B."))
        {(('A', 'B'), ('a', 'B')): 0.6, (('A', 'B'), ('a', 'b')): 0.4}

        """

        # Extended diplotypes where missing data is filled if possible
        #
        # At the moment we have decided not to implement a tailored data type
        # for diplotypes, we need to use this complicated and messy composite
        # function to try to ensure there are no duplicate permutations in the
        # filled list. If we had a hashable and symmetric data type for
        # diplotypes, we could just use set (or frozenset) to trivially reduce
        # to unique diplotypes.
        diplotypes = datautils.factorize_fill(phenotype, self.proba_haplotypes)

        # Model's haplotypes might not contain all of the constituent haplotypes
        # of the given set of phenotypes. The probability of such haplotypes
        # will be considered `fill_proba`. Note that the implied ambiguity in
        # unit summability of the NaN-probability distribution doesn't ruin the
        # calculation of diplotype probabilities (with the existing haplotypes)
        # because each conditional probability is normalized.
        proba_haplotypes = {
            h: self.proba_haplotypes.get(h, fill_proba)
            for diplotype in diplotypes for h in diplotype
        }

        def calculate(ds):
            return np.array([
                2 ** (h1 != h2) * proba_haplotypes[h1] * proba_haplotypes[h2]
                for (h1, h2) in ds
            ])

        def normalize(x):
            return x / x.sum() if any(x) else x

        def to_dict(ds):
            return dict(zip(ds, normalize(calculate(ds))))

        return to_dict(diplotypes)

    def calculate_proba_phenotypes(self, phenotype: Tuple[str], **kwargs):
        """Compute probabilities of different options that fill a given phenotype

        Examples
        --------
        >>> model = Model({("a", "b"): 0.2, ("A", "B"): 0.5, ("a", "B"): 0.3})
        >>> model.calculate_proba_phenotypes(("Aa", ".."))
        {("Aa", "BB"): 0.6, ("Aa", "Bb": 0.4)
        """
        proba_diplotypes = self.calculate_proba_diplotypes(phenotype, **kwargs)

        def agg(cum, xs):
            (diplotype, proba) = xs
            phenotype = datautils.unphase(diplotype)
            # There can be multiple different diplotypes that result into the
            # same phenotype
            return (
                {**cum, **{phenotype: proba}} if phenotype not in cum
                else {**cum, **{phenotype: proba + cum[phenotype]}}
            )

        return reduce(agg, proba_diplotypes.items(), dict())

    def impute(self, phenotype: Tuple[str], **kwargs):
        """Impute by selecting the most probable diplotype

        Examples
        --------
        >>> model = Model({("a", "b"): 0.2, ("A", "B"): 0.5, ("a", "B"): 0.3})
        >>> model.impute(("Aa", "BB"))

        """
        proba_phenotypes = self.calculate_proba_phenotypes(phenotype)
        (argmax, m) = datautils.dmax(proba_phenotypes)
        return argmax if argmax else phenotype

    def to_json(self, fp):
        """Save existing model to hard disk

        """

        def jsonify(x):
            return {"".join(k): v for (k, v) in x.items()}

        with open(fp, "w+") as f:
            json.dump(jsonify(self.proba_haplotypes), f)

        return

    @classmethod
    def from_json(cls, fp):
        """Instantiate a new model from a JSON on disk

        """

        def unjsonify(x):
            return {tuple(k): v for (k, v) in x.items()}

        with open(fp, "r") as f:
            raw = json.load(f)

        return cls(proba_haplotypes=unjsonify(raw))
